#!/usr/bin/env python
# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyscf.prop.ssc import rhf
from pyscf.prop.ssc import uhf
from pyscf.prop.ssc import dhf

from pyscf.prop.ssc.rhf import SSC as RHF
from pyscf.prop.ssc.uhf import SSC as UHF
from pyscf.prop.ssc.dhf import SSC as DHF
